"""
Top level API (:mod:`pyproject2conda`)
======================================
"""
from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as _version

try:
    __version__ = _version("pyproject2conda")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "999"

from .parser import PyProject2Conda

__author__ = """William P. Krekelberg"""
__email__ = "wpk@nist.gov"


__all__ = [
    "__version__",
    "PyProject2Conda",
]
